@@ -3,16 +3,27 @@ import dataclasses
 import itertools
 import json
 import pathlib
-import re
 import textwrap
 import uuid
 import xml.etree.ElementTree as ET
 import xml.sax.saxutils
-from typing import List, MutableMapping, Mapping, TypedDict, Set, Optional
+from typing import (
+    List,
+    MutableMapping,
+    Mapping,
+    TypedDict,
+    Set,
+    Optional,
+    Tuple,
+    TypeVar,
+    Callable,
+    Protocol,
+)
 
 import PIL
 import icontract
-import markdown
+import inflect
+import marko
 import matplotlib
 
 matplotlib.use("Agg")
@@ -218,7 +229,9 @@ def _verify_all_tags_closed(xml_text: str) -> Optional[str]:
                 return (
                     f"{exception.msg}; the line was: {line!r}, "
                     f"the open tag(s) up to that point: {list(map(_element_to_str, open_tags))}. "
-                    f"Did you maybe forget to close the tag {_element_to_str(open_tags[-1])}?"
+                    f"Did you maybe forget to close the tag {_element_to_str(open_tags[-1])}? "
+                    f"See also https://github.com/mristin/rasaeco#known-issues in case you "
+                    f"have missing or too many new lines."
                 )
             else:
                 return f"{exception.msg}; the line was: {line!r}"
@@ -250,31 +263,25 @@ def _verify_all_tags_closed(xml_text: str) -> Optional[str]:
     return None
 
 
-@dataclasses.dataclass
-class _Token:
-    identifier: str
-    text: str
-
-
 @icontract.require(lambda scenario_path: scenario_path.suffix == ".md")
-def _render_scenario(
-    scenario: rasaeco.model.Scenario,
-    ontology: rasaeco.model.Ontology,
-    scenario_path: pathlib.Path,
-    scenarios_dir: pathlib.Path,
+@icontract.require(lambda xml_path: xml_path.suffix == ".xml")
+def _render_scenario_to_xml(
+    scenario_path: pathlib.Path, xml_path: pathlib.Path
 ) -> List[str]:
-    """Render a single scenario as HTML."""
+    """Render the scenario to an intermediate XML representation."""
     try:
         text = scenario_path.read_text(encoding="utf-8")
     except Exception as exception:
-        return [f"Failed to read the scenario {scenario_path}: {exception}"]
+        return [str(exception)]
 
     ##
     # Remove <rasaeco-meta>
     ##
 
-    meta_range, errors = rasaeco.meta.find_meta(text=text)
-    assert len(errors) == 0, errors
+    meta_range, meta_errors = rasaeco.meta.find_meta(text=text)
+    if meta_errors:
+        return meta_errors
+
     assert meta_range is not None
 
     text = text[: meta_range.block_start] + text[meta_range.block_end + 1 :]
@@ -284,11 +291,9 @@ def _render_scenario(
     ##
 
     try:
-        document = markdown.markdown(text)
+        document = marko.convert(text)
     except Exception as exception:
-        return [
-            f"Failed to convert the scenario markdown {scenario_path} to HTML: {exception}"
-        ]
+        return [f"Failed to convert the scenario markdown to HTML: {exception}"]
 
     ##
     # Parse as HTML
@@ -312,91 +317,27 @@ def _render_scenario(
         ]
 
     ##
-    # Transform
+    # Perform basic validation
     ##
 
     body = root.find("body")
     assert body is not None
 
-    ##
-    # Add anchors for models
-    ##
-
-    model_set = set()  # type: Set[str]
-    in_models = False
-
-    for element in body:
-        if element.tag == "h2":
-            if in_models:
-                in_models = False
-            else:
-                if (
-                    element.text is not None
-                    and element.text.strip().lower() == "models"
-                ):
-                    in_models = True
-
-        elif in_models and element.tag == "h3":
-            name = element.text.strip() if element.text is not None else ""
-            model_set.add(name)
-
-            anchor_id = f"model-{name}"
-
-            link_el = ET.Element(
-                "a", attrib={"href": f"#{anchor_id}", "class": "anchor"}
-            )
-            link_el.append(ET.Element("a", attrib={"id": anchor_id}))
-            link_el.text = "🔗"
-            link_el.tail = element.text
-
-            element.insert(0, link_el)
-            element.text = ""
-        else:
-            pass
-
-    ##
-    # Add anchors for defs
-    ##
-
-    def_set = set()  # type: Set[str]
-    in_defs = False
-
-    for element in body:
-        if element.tag == "h2":
-            if in_defs:
-                in_defs = False
-            else:
-                if (
-                    element.text is not None
-                    and element.text.strip().lower() == "definitions"
-                ):
-                    in_defs = True
-
-        elif in_defs and element.tag == "h3":
-            name = element.text.strip() if element.text is not None else ""
-            def_set.add(name)
-
-            anchor_id = f"def-{name}"
-
-            link_el = ET.Element(
-                "a", attrib={"href": f"#{anchor_id}", "class": "anchor"}
-            )
-            link_el.append(ET.Element("a", attrib={"id": anchor_id}))
-            link_el.text = "🔗"
-            link_el.tail = element.text
-
-            element.insert(0, link_el)
-            element.text = ""
-        else:
-            pass
+    errors = []  # type: List[str]
 
     ##
     # Validate that all the tags have the "name" attribute which need to have one
     ##
 
     for element in itertools.chain(
+        root.iter("model"),
+        root.iter("def"),
+        root.iter("test"),
+        root.iter("acceptance"),
         root.iter("ref"),
         root.iter("modelref"),
+        root.iter("testref"),
+        root.iter("acceptanceref"),
         root.iter("phase"),
         root.iter("level"),
     ):
@@ -408,52 +349,290 @@ def _render_scenario(
     if errors:
         return errors
 
-    ##
-    # Validate that all modelrefs have the model defined
-    ##
+    try:
+        xml_path.write_text(html_text)
+    except Exception as error:
+        return [
+            f"Failed to store the intermediate XML representation "
+            f"of a scenario {scenario_path} to {xml_path}: {error}"
+        ]
 
-    for element in root.iter("modelref"):
-        modelref = element.attrib["name"]
-        if modelref not in model_set:
-            errors.append(f"The modelref is invalid: {modelref!r}")
+    return []
 
-    ##
-    # Validate that all refs have the def defined
-    ##
 
-    for element in root.iter("ref"):
-        ref = element.attrib["name"]
-        if ref not in def_set:
-            errors.append(f"The ref is invalid: {ref!r}")
+@icontract.require(lambda xml_path: xml_path.suffix == ".xml")
+def _extract_definitions(
+    xml_path: pathlib.Path,
+) -> Tuple[Optional[rasaeco.model.Definitions], List[str]]:
+    """
+    Extract the definitions from the intermediate representation of a scenario.
+
+    Return (definitions, errors if any).
+    """
+    try:
+        text = xml_path.read_text(encoding="utf-8")
+    except Exception as exception:
+        return None, [
+            f"Failed to read the intermediate representation "
+            f"of the scenario {xml_path}: {exception}"
+        ]
+
+    root = ET.fromstring(text)
+
+    def collect_set_of_named_references(tag: str) -> Set[str]:
+        """Collect the set of references for the given specification tag."""
+        result = set()  # type: Set[str]
+        for element in root.iter(tag):
+            name = element.attrib["name"]
+            result.add(name)
+        return result
+
+    return (
+        rasaeco.model.Definitions(
+            model_set=collect_set_of_named_references(tag="model"),
+            def_set=collect_set_of_named_references(tag="def"),
+            test_set=collect_set_of_named_references(tag="test"),
+            acceptance_set=collect_set_of_named_references(tag="acceptance"),
+        ),
+        [],
+    )
+
+
+@icontract.require(
+    lambda element: element.tag in ["ref", "modelref", "testref", "acceptanceref"]
+)
+def _parse_reference_element(element: ET.Element) -> Tuple[Optional[str], str]:
+    """Extract the scenario identifier and the name from a reference element."""
+    name_attribute = element.attrib["name"]
+    if "#" in name_attribute:
+        scenario_id, name = name_attribute.split("#")
+        return scenario_id, name
+    else:
+        return None, name_attribute
+
+
+@icontract.require(lambda xml_path: xml_path.suffix == ".xml")
+def _validate_references(
+    scenario: rasaeco.model.Scenario,
+    ontology: rasaeco.model.Ontology,
+    xml_path: pathlib.Path,
+) -> List[str]:
+    """Validate that all the references are valid in the given scenario."""
+    try:
+        text = xml_path.read_text(encoding="utf-8")
+    except Exception as exception:
+        return [
+            f"Failed to read the intermediate representation "
+            f"of the scenario {xml_path}: {exception}"
+        ]
+
+    root = ET.fromstring(text)
+
+    body = root.find("body")
+    assert body is not None
+
+    class SetGetterForScenario(Protocol):
+        def __call__(self, scenario_id: str) -> Set[str]:
+            ...
+
+    @icontract.require(
+        lambda reference_tag: reference_tag
+        in ["modelref", "ref", "testref", "acceptanceref"]
+    )
+    def validate_references_for_tag(
+        reference_tag: str, set_getter_for_scenario: SetGetterForScenario
+    ) -> List[str]:
+        """Validate that the reference tags refer to the actual definitions."""
+        errors = []  # type: List[str]
+
+        for element in root.iter(reference_tag):
+            scenario_id, name = _parse_reference_element(element=element)
+            scenario_id = (
+                scenario_id if scenario_id is not None else scenario.identifier
+            )
+
+            if scenario_id not in ontology.scenario_map:
+                errors.append(
+                    f"The {reference_tag} is invalid: {_element_to_str(element)}; "
+                    f"the scenario with the identifier {scenario_id} does not exist."
+                )
+            elif name not in set_getter_for_scenario(scenario_id=scenario_id):
+                errors.append(
+                    f"The {reference_tag} is invalid: {_element_to_str(element)!r}; "
+                    f"the specified target {name!r} is missing in the scenario {scenario_id}."
+                )
+            else:
+                # The reference is valid.
+                pass
 
-    if errors:
         return errors
 
+    return (
+        validate_references_for_tag(
+            reference_tag="modelref",
+            set_getter_for_scenario=lambda scenario_id: ontology.scenario_map[
+                scenario_id
+            ].definitions.model_set,
+        )
+        + validate_references_for_tag(
+            reference_tag="ref",
+            set_getter_for_scenario=lambda scenario_id: ontology.scenario_map[
+                scenario_id
+            ].definitions.def_set,
+        )
+        + validate_references_for_tag(
+            reference_tag="testref",
+            set_getter_for_scenario=lambda scenario_id: ontology.scenario_map[
+                scenario_id
+            ].definitions.test_set,
+        )
+        + validate_references_for_tag(
+            reference_tag="acceptanceref",
+            set_getter_for_scenario=lambda scenario_id: ontology.scenario_map[
+                scenario_id
+            ].definitions.acceptance_set,
+        )
+    )
+
+
+_INFLECT_ENGINE = inflect.engine()
+
+
+@icontract.require(lambda xml_path: xml_path.suffix == ".xml")
+def _render_scenario(
+    scenario: rasaeco.model.Scenario,
+    ontology: rasaeco.model.Ontology,
+    xml_path: pathlib.Path,
+    html_path: pathlib.Path,
+) -> List[str]:
+    """Render a single scenario as HTML."""
+    try:
+        text = xml_path.read_text(encoding="utf-8")
+    except Exception as exception:
+        return [
+            f"Failed to read the intermediate representation "
+            f"of the scenario {xml_path}: {exception}"
+        ]
+
+    rel_pth_to_scenario_dir = pathlib.PurePosixPath(
+        *([".."] * len(scenario.relative_path.parent.parts))
+    )
+
+    root = ET.fromstring(text)
+
+    body = root.find("body")
+    assert body is not None
+
     ##
-    # Replace <ref> tags with proper HTML
+    # Convert specification tags to proper HTML
     ##
 
+    def convert_tags_to_html(tag: str, readable_title: bool) -> None:
+        """Convert a specification tag, such as <model> to proper HTML."""
+        for element in root.iter(tag):
+            name = element.attrib["name"]
+
+            element.tag = "div"
+            element.attrib = {"class": tag}
+
+            header_el = ET.Element("h3")
+
+            anchor_el = ET.Element("a")
+            anchor_el.attrib = {"name": f"{tag}-{name}"}
+            anchor_el.text = " "
+            header_el.insert(0, anchor_el)
+
+            link_el = ET.Element("a")
+            link_el.attrib = {"href": f"#{tag}-{name}", "class": "anchor"}
+            link_el.text = "🔗"
+
+            if readable_title:
+                link_el.tail = name.replace("_", " ")
+            else:
+                link_el.tail = name
+
+            header_el.insert(0, link_el)
+            header_el.tail = "\n"
+
+            element.insert(0, header_el)
+
+    convert_tags_to_html(tag="model", readable_title=False)
+    convert_tags_to_html(tag="def", readable_title=True)
+    convert_tags_to_html(tag="test", readable_title=True)
+    convert_tags_to_html(tag="acceptance", readable_title=True)
+
+    ##
+    # Convert references to proper HTML
+    ##
+
+    # <ref> is a special case as we need to pluralize and prettify.
     for element in root.iter("ref"):
-        anchor_name = element.attrib["name"]
+        scenario_id, ref = _parse_reference_element(element=element)
+
+        readable = ref.replace("_", " ")
+        if element.tail is not None and element.tail.startswith("s"):
+            element.tail = element.tail[1:]
+            readable = _INFLECT_ENGINE.plural_noun(readable)
+
+        if scenario_id is None:
+            link_text = readable
+            href = f"#def-{ref}"
+        else:
+            link_text = f"{readable} (from {scenario_id})"
+
+            href_pth = _html_path(
+                scenario_path=rel_pth_to_scenario_dir
+                / ontology.scenario_map[scenario_id].relative_path
+            )
+            href = f"{href_pth.as_posix()}#def-{ref}"
 
         element.tag = "a"
-        element.attrib = {"href": f"#def-{anchor_name}", "class": "ref"}
+        element.attrib = {"href": href, "class": "ref"}
 
         if len(element) == 0 and not element.text:
-            element.text = anchor_name
+            element.text = link_text
 
-    ##
-    # Replace <modelref> tags with proper HTML
-    ##
+    @icontract.require(
+        lambda reference_tag: reference_tag
+        in ["modelref", "ref", "testref", "acceptanceref"]
+    )
+    def convert_references_to_html(reference_tag: str) -> None:
+        """Convert the reference tags to proper HTML."""
+        if reference_tag == "modelref":
+            target_tag = "model"
+        elif reference_tag == "ref":
+            target_tag = "def"
+        elif reference_tag == "testref":
+            target_tag = "test"
+        elif reference_tag == "acceptanceref":
+            target_tag = "acceptance"
+        else:
+            raise ValueError(f"Unexpected reference tag: {reference_tag!r}")
 
-    for element in root.iter("modelref"):
-        anchor_name = element.attrib["name"]
+        for element in root.iter(reference_tag):
+            scenario_id, name = _parse_reference_element(element=element)
 
-        element.tag = "a"
-        element.attrib = {"href": f"#model-{anchor_name}", "class": "modelref"}
+            if scenario_id is None:
+                link_text = name
+                href = f"#{target_tag}-{name}"
+            else:
+                link_text = f"{name} (from {scenario_id})"
 
-        if len(element) == 0 and not element.text:
-            element.text = anchor_name
+                href_pth = _html_path(
+                    scenario_path=rel_pth_to_scenario_dir
+                    / ontology.scenario_map[scenario_id].relative_path
+                )
+                href = f"{href_pth.as_posix()}#{target_tag}-{name}"
+
+            element.tag = "a"
+            element.attrib = {"href": href, "class": reference_tag}
+
+            if len(element) == 0 and not element.text:
+                element.text = link_text
+
+    convert_references_to_html(reference_tag="modelref")
+    convert_references_to_html(reference_tag="testref")
+    convert_references_to_html(reference_tag="acceptanceref")
 
     ##
     # Replace <phase> tags with proper HTML
@@ -468,12 +647,13 @@ def _render_scenario(
 
     for element in root.iter("phase"):
         name = element.attrib["name"]
+        readable = name.replace("_", " ")
 
         element.tag = "span"
         element.attrib = {"class": "phase", "data-text": name}
 
         sup_el = ET.Element("sup")
-        sup_el.text = name
+        sup_el.text = readable
         element.append(sup_el)
 
         anchor = f"phase-anchor-{uuid.uuid4()}"
@@ -494,12 +674,13 @@ def _render_scenario(
 
     for element in root.iter("level"):
         name = element.attrib["name"]
+        readable = name.replace("_", " ")
 
         element.tag = "span"
         element.attrib = {"class": "level", "data-text": name}
 
         sup_el = ET.Element("sup")
-        sup_el.text = name
+        sup_el.text = readable
         element.append(sup_el)
 
         anchor = f"level-anchor-{uuid.uuid4()}"
@@ -690,6 +871,12 @@ def _render_scenario(
 
     body.insert(0, img)
 
+    ##
+    # Insert the contact
+    ##
+
+    body.insert(0, _element_with_text(tag="p", text=scenario.contact))
+
     ##
     # Insert the title
     ##
@@ -703,9 +890,7 @@ def _render_scenario(
     back_link = ET.Element("a")
 
     back_url = (
-        pathlib.PurePosixPath(
-            *([".."] * len(scenario_path.parent.relative_to(scenarios_dir).parts))
-        )
+        pathlib.PurePosixPath(*([".."] * len(scenario.relative_path.parent.parts)))
         / "ontology.html"
     ).as_posix()
 
@@ -717,16 +902,27 @@ def _render_scenario(
     # Save
     ##
 
-    target_pth = scenario_path.parent / (scenario_path.stem + ".html")
-
     try:
-        target_pth.write_bytes(ET.tostring(root, encoding="utf-8"))
+        html_path.write_bytes(ET.tostring(root, encoding="utf-8"))
     except Exception as exception:
-        return [f"Failed to write generated HTML code to {target_pth}: {exception}"]
+        return [f"Failed to write generated HTML code to {html_path}: {exception}"]
 
     return []
 
 
+def _xml_path(scenario_path: pathlib.Path) -> pathlib.Path:
+    """Generate the corresponding XML path of the intermediate representation."""
+    return scenario_path.parent / (scenario_path.stem + ".xml")
+
+
+PathT = TypeVar("PathT", pathlib.Path, pathlib.PurePosixPath)
+
+
+def _html_path(scenario_path: PathT) -> PathT:
+    """Generate the corresponding path of the HTML representation."""
+    return scenario_path.parent / (scenario_path.stem + ".html")
+
+
 def once(scenarios_dir: pathlib.Path) -> List[str]:
     """
     Render the scenarios and the ontology.
@@ -737,7 +933,24 @@ def once(scenarios_dir: pathlib.Path) -> List[str]:
     meta_map = dict()  # type: MutableMapping[str, rasaeco.meta.Meta]
 
     errors = []  # type: List[str]
-    for pth in sorted(scenarios_dir.glob("**/*.md")):
+
+    scenario_pths = sorted(scenarios_dir.glob("**/scenario.md"))
+    for pth in scenario_pths:
+        to_xml_errors = _render_scenario_to_xml(
+            scenario_path=pth, xml_path=_xml_path(pth)
+        )
+        for error in to_xml_errors:
+            errors.append(
+                f"When rendering {pth} to intermediate XML representation: {error}"
+            )
+
+        if errors:
+            continue
+
+    if errors:
+        return errors
+
+    for pth in scenario_pths:
         meta, meta_errors = rasaeco.meta.extract_meta(
             text=pth.read_text(encoding="utf-8")
         )
@@ -750,11 +963,6 @@ def once(scenarios_dir: pathlib.Path) -> List[str]:
 
         assert meta is not None
 
-        if meta["identifier"] in path_map:
-            errors.append(
-                f"In file {pth}: Identifier conflicts with the file {path_map[meta['identifier']]}"
-            )
-
         for i, cubelet in enumerate(meta["volumetric"]):
             ##
             # Verify aspect range
@@ -787,8 +995,10 @@ def once(scenarios_dir: pathlib.Path) -> List[str]:
                     f"In file {pth} and cubelet {i + 1}: Invalid level range: {range_error}"
                 )
 
-        meta_map[meta["identifier"]] = meta
-        path_map[meta["identifier"]] = pth
+        identifier = pth.parent.relative_to(scenarios_dir).as_posix()
+
+        meta_map[identifier] = meta
+        path_map[identifier] = pth
 
     scenario_id_set = set(meta_map.keys())
 
@@ -822,14 +1032,23 @@ def once(scenarios_dir: pathlib.Path) -> List[str]:
                 )
             )
 
-        scenario = rasaeco.model.Scenario(
-            identifier=identifier,
-            title=meta["title"],
-            volumetric=volumetric,
-            relative_path=path_map[identifier].relative_to(scenarios_dir),
-        )
+        pth = path_map[identifier]
+        definitions, extraction_errors = _extract_definitions(xml_path=_xml_path(pth))
+        if extraction_errors:
+            errors.extend(extraction_errors)
+        else:
+            assert definitions is not None
 
-        scenarios.append(scenario)
+            scenario = rasaeco.model.Scenario(
+                identifier=identifier,
+                title=meta["title"],
+                contact=meta["contact"],
+                volumetric=volumetric,
+                definitions=definitions,
+                relative_path=pth.relative_to(scenarios_dir),
+            )
+
+            scenarios.append(scenario)
 
     relations = []  # type: List[rasaeco.model.Relation]
     for identifier, meta in meta_map.items():
@@ -852,11 +1071,24 @@ def once(scenarios_dir: pathlib.Path) -> List[str]:
 
     for scenario in scenarios:
         pth = path_map[scenario.identifier]
+        validation_errors = _validate_references(
+            scenario=scenario, ontology=ontology, xml_path=_xml_path(pth)
+        )
+
+        for error in validation_errors:
+            errors.append(f"When validating references in {pth}: {error}")
+
+    if errors:
+        return errors
+
+    for scenario in scenarios:
+        pth = path_map[scenario.identifier]
+
         render_errors = _render_scenario(
             scenario=scenario,
             ontology=ontology,
-            scenario_path=pth,
-            scenarios_dir=scenarios_dir,
+            xml_path=_xml_path(pth),
+            html_path=_html_path(pth),
         )
 
         for error in render_errors:
